#ifndef NRLIB_PCHHEADER_HPP
#define NRLIB_PCHHEADER_HPP

#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif
#include <cmath>

#include <string>
#include <iostream>
#include <vector>
#include <map>

#include <boost/filesystem.hpp>

#endif
