// $Id: generaldiscrete.cpp 1081 2012-09-28 09:11:07Z perroe $

// Copyright (c)  2011, Norwegian Computing Center
// All rights reserved.
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// •  Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
// •  Redistributions in binary form must reproduce the above copyright notice, this list of
//    conditions and the following disclaimer in the documentation and/or other materials
//    provided with the distribution.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
// SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
// OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "generaldiscrete.hpp"

namespace NRLib {

GeneralDiscrete::GeneralDiscrete(const std::vector<double> & hist)
  : hist_(hist)
{
  NormalizeHist();
  MakeCumHist();
}

GeneralDiscrete::~GeneralDiscrete()
{}

Distribution<int>*
GeneralDiscrete::Clone() const
{
  return new GeneralDiscrete(*this);
}

void
GeneralDiscrete::NormalizeHist()
{
  double sum(0.0);
  for (size_t i = 0; i < hist_.size(); i++)
    sum += hist_[i];
  for (size_t i = 0; i < hist_.size(); i++)
    hist_[i] /= sum;
}

void
GeneralDiscrete::MakeCumHist()
{
  double sum(0.0);
  cum_hist_.resize(hist_.size());
  for (size_t i = 0; i < hist_.size(); i++) {
    sum += hist_[i];
    cum_hist_[i] = sum;
  }
}

} // namespace NRLib
