// $Id: generaldiscrete.hpp 1468 2017-04-21 13:09:33Z perroe $

// Copyright (c)  2011, Norwegian Computing Center
// All rights reserved.
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// •  Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
// •  Redistributions in binary form must reproduce the above copyright notice, this list of
//    conditions and the following disclaimer in the documentation and/or other materials
//    provided with the distribution.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
// SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
// OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef NRLIB_RANDOM_GENERALDISCRETE_HPP
#define NRLIB_RANDOM_GENERALDISCRETE_HPP

#include "distribution.hpp"

#include <cassert>
#include <vector>

namespace NRLib {

class GeneralDiscrete : public Distribution<int> {
public:
  explicit GeneralDiscrete(const std::vector<double> & hist = std::vector<double>(0));
  ~GeneralDiscrete();

  Distribution<int>* Clone() const;

  inline double Pdf(int i) const { return hist_[i]; }
  inline double Cdf(int i) const { return cum_hist_[i]; }
  inline int Quantile(double x) const;

  size_t GetN() const { return hist_.size(); }

private:
  void NormalizeHist();
  void MakeCumHist();

  std::vector<double> hist_;
  std::vector<double> cum_hist_;
};

// ---------------- IMPLEMENTATION OF INLINE FUNCTIONS -----------------

int GeneralDiscrete::Quantile(double p) const {
  assert(p >= 0.0 && p <= 1.0);
  int i = 0;
  while (cum_hist_[i] < p) {
    i++;
  }
  return(i);
}

}

#endif
