/*
 *   Copyright (c) 2007, Michael Lehn
 *
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1) Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2) Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *   3) Neither the name of the FLENS development group nor the names of
 *      its contributors may be used to endorse or promote products derived
 *      from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef FLENS_LAPACK_FLENS_AUX_H
#define FLENS_LAPACK_FLENS_AUX_H 1

#include <flens/densevector.h>
#include <flens/generalmatrix.h>
#include <flens/lapack_flens.h>

namespace flens {

template <typename Mat, typename Vec>
class OrthMatrixQR;

template <typename Mat, typename Vec>
struct TypeInfo<OrthMatrixQR<Mat,Vec> >
{
    typedef OrthMatrixQR<Mat, Vec>  Impl;
    typedef double                  ElementType;
};

template <typename Mat, typename Vec>
class OrthMatrixQR
    : public GeneralMatrix<OrthMatrixQR<Mat, Vec> >
{
    public:
        OrthMatrixQR(const Mat &_QR, const Vec &_tau)
            : QR(_QR), tau(_tau)
        {
        }

        const Mat &QR;
        const Vec &tau;
};

template <typename ALPHA, typename M, typename V, typename Mat, typename BETA>
void
mm(Transpose transA, Transpose transB, ALPHA alpha,
   const OrthMatrixQR<M,V> &A, const Mat &X,
   BETA beta, Mat &B);

template <typename ALPHA, typename M, typename V, typename Mat, typename BETA>
void
mm(Transpose transB, Transpose transA, ALPHA alpha,
   const Mat &X, const OrthMatrixQR<M,V> &A,
   BETA beta, Mat &B);

} // namespace flens

#include <flens/lapack_flens_aux.tcc>

#endif // FLENS_LAPACK_FLENS_AUX_H
